"""Abstract Keras Sequence object for running models on huge datasets."""
from keras_mixed_sequence import Sequence


class AbstractSequence(Sequence):
    """Abstract Keras Sequence object for running models on huge datasets."""

    def __init__(
        self,
        batch_size: int,
        samples_number: int,
        window_size: int = 4,
        shuffle: bool = True,
        elapsed_epochs: int = 0,
    ):
        """Create new Sequence object.

        Parameters
        -----------------------------
        batch_size: int,
            Number of nodes to include in a single batch.
        samples_number: int,
            Number of samples that compose this Sequence.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        shuffle: bool = True,
            Wthever to shuffle the vectors.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        """
        self._window_size = window_size
        self._shuffle = shuffle

        super().__init__(
            samples_number=samples_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )



"""EdgeTransformer class to convert edges to edge embeddings."""
import numpy as np
from .node_transformer import NodeTransformer


class EdgeTransformer:
    """EdgeTransformer class to convert edges to edge embeddings."""

    methods = {
        "hadamard": np.multiply,
        "average": lambda x1, x2: np.mean([x1, x2], axis=0),
        "weightedL1": lambda x1, x2: np.abs(x1 - x2),
        "weightedL2": lambda x1, x2: (x1 - x2)**2
    }

    def __init__(self, method: str = "hadamard"):
        """Create new EdgeTransformer object.

        Parameters
        ------------------------
        method: str = "hadamard",
            Method to use for the embedding.
            Can either be 'hadamard', 'average', 'weightedL1', 'weightedL2' or
            a custom lambda that receives two numpy arrays with the nodes
            embedding and returns the edge embedding.
        """
        if isinstance(method, str) and method not in EdgeTransformer.methods:
            raise ValueError((
                "Given method '{}' is not supported. "
                "Supported methods are {}, or alternatively a lambda."
            ).format(
                method, ", ".join(list(EdgeTransformer.methods.keys()))
            ))
        self._method = EdgeTransformer.methods[method]
        self._transformer = NodeTransformer()

    def fit(self, embedding: np.ndarray):
        """Fit the model.

        Parameters
        -------------------------
        embedding: np.ndarray,
            Embedding to use to fit the transformer.
        """
        self._transformer.fit(embedding)

    def transform(self, sources: np.ndarray, destinations: np.ndarray) -> np.ndarray:
        """Return embedding for given edges using provided method.

        Parameters
        --------------------------
        sources: np.ndarray,
            Vector of source nodes whose embedding is to be returned.
        destinations:np.ndarray,
            Vector of destination nodes whose embedding is to be returned.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        return self._method(
            self._transformer.transform(sources),
            self._transformer.transform(destinations),
        )



import collections
import math
import numpy as np  # type: ignore
import random
import tensorflow as tf  # type: ignore

from tqdm import trange  # type: ignore
from typing import Dict, List, Optional, Tuple, Union

from embiggen.utils import get_embedding, calculate_cosine_similarity
import logging


class Word2Vec:
    """Superclass of all of the word2vec family algorithms.
        Attributes:
            learning_rate: A float between 0 and 1 that controls how fast the model learns to solve the problem.
            batch_size: The size of each "batch" or slice of the data to sample when training the model.
            num_epochs: The number of epochs to run when training the model.
            display_step: An integer that is used to determine the number of steps to display.
            eval_step: This attribute stores the total number of iterations to run during training.
            embedding_size: Dimension of embedded vectors.
            max_vocabulary_size: Maximum number of words (i.e. total number of different words in the vocabulary).
            vocabulary_size: An integer storing the total number of unique words in the vocabulary.
            min_occurrence: Minimum number of times a word needs to appear to be included (default=1).
            context_window: How many words to consider left and right.
            num_skips: How many times to reuse an input to generate a label.
            num_sampled: Number of negative examples to sample (default=7).
            display: An integer of the number of words to display.
        """

    def __init__(self, data: List, worddictionary: Dict[str, int], reverse_worddictionary: Dict[int, str],
                 learning_rate: float = 0.1, batch_size: int = 128,
                 num_epochs: int = 1, embedding_size: int = 200, max_vocabulary_size: int = 50000,
                 min_occurrence: int = 1, context_window: int = 3, num_skips: int = 2, num_sampled: int = 7,
                 display: Optional[int] = None, device_type: str = 'cpu') -> None:
        self.data = data
        self.word2id = worddictionary
        self.id2word = reverse_worddictionary
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.n_epochs = num_epochs
        self.display_step = 2000
        self.eval_step = 100
        self.embedding_size = embedding_size
        self.max_vocabulary_size = max_vocabulary_size
        self.vocabulary_size: int = 0
        self.min_occurrence = min_occurrence
        self.context_window = context_window
        self.num_skips = num_skips
        self.num_sampled = num_sampled
        self.display = display
        self.display_examples: List = []
        self.device_type = '/CPU:0' if 'cpu' in device_type.lower() else '/GPU:0'

        # takes the input data and goes through each element
        # first, check each element is a list
        # if any(isinstance(el, list) for el in self.data):
        if isinstance(self.data, tf.RaggedTensor):
            self.list_of_lists = True
        elif isinstance(self.data, tf.Tensor):
            self.list_of_lists = False
        else:
            logging.info("NEITHER RAGGED NOR TENSOR")
            logging.info("Type of data:{} ".format(type(self.data)))
            raise TypeError("NEITHER RAGGED NOR TENSOR")

    def add_display_words(self, count: list, num: int = 5) -> None:
        """Creates a list of display nodes/words by obtaining a random sample of 'num' nodes/words from the full
        sample.

        If the argument 'display' is not None, then we expect that the user has passed an integer amount of words
        that are to be displayed together with their nearest neighbors, as defined by the word2vec algorithm. It is
        important to note that display is a costly operation. Up to 16 nodes/words are permitted. If a user asks for
        more than 16, a random validation set of 'num' nodes/words, that includes common and uncommon nodes/words, is
        selected from a 'valid_window' of 50 nodes/words.
        Args:
            count: A list of tuples (key:word, value:int).
            num: An integer representing the number of words to sample.
        Returns:
            None.
        Raises:
            TypeError: If the user does not provide a list of display words.
        """

        if not isinstance(count, list):
            self.display = None
            raise TypeError('self.display requires a list of tuples with key:word, value:int (count)')

        if num > 16:
            logging.warning('maximum of 16 display words allowed (you passed {num_words})'.format(num_words=num))
            num = 16

        # pick a random validation set of 'num' words to sample
        valid_window = 50
        valid_examples = np.array(random.sample(range(2, valid_window), num))

        # sample less common words - choose 'num' points randomly from the first 'valid_window' after element 1000
        self.display_examples = np.append(valid_examples, random.sample(range(1000, 1000 + valid_window), num), axis=0)

        return None

    def calculate_vocabulary_size(self) -> None:
        """Calculates the vocabulary size for the input data, which is a list of words (i.e. from a text),
        or list of lists (i.e. from a collection of sentences or random walks).
        The function checks that self.vocabulary size has not been set

        Returns:
            None.
        """
        if self.word2id is None and self.vocabulary_size == 0:
            self.vocabulary_size = 0
        elif self.vocabulary_size ==0:
            self.vocabulary_size = len(self.word2id)+1
        # Apr 28, changed by Peter (can be deleted)
        # if any(isinstance(el, list) for el in data):
        #    flat_list = [item for sublist in data for item in sublist]
        #    self.vocabulary_size = min(self.max_vocabulary_size, len(set(flat_list)) + 1)
        #    print('Vocabulary size (list of lists) is {vocab_size}'.format(vocab_size=self.vocabulary_size))
        # else:
        ##    # was - self.vocabulary_size = min(self.max_vocabulary_size, len(set(data)) + 1)
        #    self.vocabulary_size = min(self.max_vocabulary_size, TFUtilities.gets_tensor_length(data) + 1)
        # print('Vocabulary size (flat) is {vocab_size}'.format(vocab_size=self.vocabulary_size))
        return None


class SkipGramWord2Vec(Word2Vec):
    """
    Class to run word2vec using skip grams
    """

    def __init__(self, data: List, worddictionary: Dict[str, int], reverse_worddictionary: Dict[int, str],
                 learning_rate: float = 0.1, batch_size: int = 128, num_epochs: int = 1, embedding_size: int = 200,
                 max_vocabulary_size: int = 50000, min_occurrence: int = 1, context_window: int = 3, num_skips: int = 2,
                 num_sampled: int = 7, display: Optional[int] = None, device_type: str = 'cpu') -> None:

        super().__init__(data=data, worddictionary=worddictionary, reverse_worddictionary=reverse_worddictionary,
                         learning_rate=learning_rate, batch_size=batch_size, num_epochs=num_epochs,
                         embedding_size=embedding_size, max_vocabulary_size=max_vocabulary_size,
                         min_occurrence=min_occurrence, context_window=context_window, num_skips=num_skips,
                         num_sampled=num_sampled, display=display, device_type=device_type)

        self.data = data
        self.device_type = '/CPU:0' if 'cpu' in device_type.lower() else '/GPU:0'
        # set vocabulary size
        self.calculate_vocabulary_size()

        # with toy exs the # of nodes might be lower than the default value of num_sampled of 7. num_sampled needs to
        # be less than the # of exs (num_sampled is the # of negative samples that get evaluated per positive ex)
        if self.num_sampled > self.vocabulary_size:
            self.num_sampled = int(self.vocabulary_size / 2)

        self.optimizer: tf.keras.optimizers = tf.keras.optimizers.SGD(learning_rate)
        self.data_index: int = 0
        self.current_sentence: int = 0

        # do not display examples during training unless the user calls add_display_words (i.e. default is None)
        self.display = display
        self.n_epochs = num_epochs

        # ensure the following ops & var are assigned on CPU (some ops are not compatible on GPU)
        with tf.device(self.device_type):
            # create embedding (each row is a word embedding vector) with shape (#n_words, dims) and dim = vector size
            self.embedding: tf.Variable = tf.Variable(tf.random.normal([self.vocabulary_size, embedding_size]))

            # construct the variables for the NCE loss
            self.nce_weights: tf.Variable = tf.Variable(tf.random.normal([self.vocabulary_size, embedding_size]))
            self.nce_biases: tf.Variable = tf.Variable(tf.zeros([self.vocabulary_size]))

    def nce_loss(self, x_embed: tf.Tensor, y: np.ndarray) -> Union[float, int]:
        """Calculates the noise-contrastive estimation (NCE) training loss estimation for each batch.
        Args:
            x_embed: A Tensor with shape [batch_size, dim].
            y: An array containing the target classes with shape [batch_size, num_true].
        Returns:
            loss: The NCE losses.
        """

        with tf.device(self.device_type):
            y = tf.cast(y, tf.int64)

            loss = tf.reduce_mean(tf.nn.nce_loss(weights=self.nce_weights,
                                                 biases=self.nce_biases,
                                                 labels=y,
                                                 inputs=x_embed,
                                                 num_sampled=self.num_sampled,
                                                 num_classes=self.vocabulary_size)
                                  )

            return loss

    # def evaluate(self, x_embed: tf.Tensor) -> Union[np.ndarray, tf.Tensor]:
    #     """Computes the cosine similarity between a provided embedding and all other embedding vectors.
    #
    #     Args:
    #         x_embed: A Tensor containing word embeddings.
    #
    #     Returns:
    #         cosine_sim_op: A tensor of the cosine similarities between input data embedding and all other embeddings.
    #     """
    #
    #     with tf.device(self.device_type):
    #         x_embed_cast = tf.cast(x_embed, tf.float32)
    #         x_embed_norm = x_embed_cast / tf.sqrt(tf.reduce_sum(tf.square(x_embed_cast)))
    #         x_embed_sqrt = tf.sqrt(tf.reduce_sum(tf.square(self.embedding), 1, keepdims=True), tf.float32)
    #         embedding_norm = self.embedding / x_embed_sqrt
    #
    #         # calculate cosine similarity
    #         cosine_sim_op = tf.matmul(x_embed_norm, embedding_norm, transpose_b=True)
    #
    #         return cosine_sim_op

    def run_optimization(self, x: np.array, y: np.array) -> float:
        """Runs optimization for each batch by retrieving an embedding and calculating loss. Once the loss has been
        calculated, the gradients are computed and the weights and biases are updated accordingly.
        Args:
            x: An array of integers to use as batch training data.
            y: An array of labels to use when evaluating loss for an epoch.
        Returns:
            The loss of the current optimization round.
        """

        with tf.device(self.device_type):
            # wrap computation inside a GradientTape for automatic differentiation
            with tf.GradientTape() as g:
                embedding = get_embedding(x, self.embedding, self.device_type)
                loss = self.nce_loss(embedding, y)

            # compute gradients
            gradients = g.gradient(loss, [self.embedding, self.nce_weights, self.nce_biases])

            # update W and b following gradients
            self.optimizer.apply_gradients(zip(gradients, [self.embedding, self.nce_weights, self.nce_biases]))

        return loss

    def next_batch(self, sentence: tf.Tensor) -> Tuple[np.ndarray, np.ndarray]:
        """Generate training batch for the skip-gram model.

        Assumption: This assumes that dslist is a td.data.Dataset object that contains one sentence or (or list of words

        Args:
            sentence: A list of words to be used to create the batch
        Returns:
            A list where the first item us a batch and the second item is the batch's labels.
        Raises:
            ValueError: If the number of skips is not <= twice the skip window length.

        TODO -- should num_skips and context_window be arguments or simply taken from self
        within the method?
            """
        num_skips = self.num_skips
        context_window = self.context_window
        if num_skips > 2 * context_window:
            raise ValueError('The value of self.num_skips must be <= twice the length of self.context_window')
        # TODO  -- We actually only need to check the above once in the Constructor?
        # OR -- is there any situation where we will change this during training??
        # self.data is a list of lists, e.g., [[1, 2, 3], [5, 6, 7]]
        span = 2 * context_window + 1
        # again, probably we can go: span = self.span
        sentencelen = len(sentence)
        batch_size = ((sentencelen - (2 * context_window)) * num_skips)
        batch = np.ndarray(shape=(batch_size,), dtype=np.int32)
        labels = np.ndarray(shape=(batch_size, 1), dtype=np.int32)
        buffer: collections.deque = collections.deque(maxlen=span)
        sentence = sentence.numpy()
        # The following command fills up the Buffer but leaves out the last spot
        # this allows us to always add the next word as the first thing we do in the
        # following loop.
        buffer.extend(sentence[0:span - 1])
        data_index = span - 1
        for i in range(batch_size // num_skips):
            buffer.append(sentence[data_index])
            data_index += 1  # move sliding window 1 spot to the right
            context_words = [w for w in range(span) if w != context_window]
            words_to_use = random.sample(context_words, num_skips)

            for j, context_word in enumerate(words_to_use):
                batch[i * num_skips + j] = buffer[context_window]
                labels[i * num_skips + j, 0] = buffer[context_word]
        return batch, labels

    def display_words(self, x_test: np.array) -> None:
        """
        This is intended for to give feedback on the shell about the progress of training.
        It is not needed for the actual analysis.
        :param x_test:
        :return:
        """
        logging.info("Evaluation...")
        sim = calculate_cosine_similarity(get_embedding(x_test, self.embedding, self.device_type),
                                          self.embedding,
                                          self.device_type).numpy()
        #print(sim[0])
        for i in range(len(self.display_examples)):
            top_k = 8  # number of nearest neighbors.
            nearest = (-sim[i, :]).argsort()[1:top_k + 1]
            disp_example = self.id2word[self.display_examples[i]]
            log_str = '"%s" nearest neighbors:' % disp_example
            for k in range(top_k):
                log_str = '%s %s,' % (log_str, self.id2word[nearest[k]])
            #print(log_str)

    def train(self) -> List[float]:
        """
        Trying out passing a simple Tensor to get_batch
        :return:
        """
        # words for testing; display_step = 2000; eval_step = 2000
        do_display = self.display_step is not None and len(self.display_examples) > 0

        if do_display:
            for w in self.display_examples:
                logging.info('{word}: id={index}'.format(word=self.id2word[w], index=w))

        x_test = np.array(self.display_examples)

        window_len = 2 * self.context_window + 1
        step = 0
        loss_history = []
        for _ in trange(1, self.n_epochs + 1):
            if self.list_of_lists or isinstance(self.data, tf.RaggedTensor):
                for sentence in self.data:
                    # Sentence is a Tensor
                    sentencelen = len(sentence)
                    if sentencelen < window_len:
                        continue
                    batch_x, batch_y = self.next_batch(sentence)
                    # Evaluation.
                    if do_display and (step % self.eval_step == 0 or step == 0):
                        self.display_words(x_test)
                    step += 1
                    self.run_optimization(batch_x, batch_y)
            else:
                data = self.data  #
                if not isinstance(data, tf.Tensor):
                    raise TypeError("We were expecting a Tensor object!")
                # batch_size = self.batch_size
                data_len = len(data)
                # Note that we cannot fully digest all of the data in any one batch
                # if the window length is K and the natch_len is N, then the last
                # window that we get starts at position (N-K). Therefore, if we start
                # the next window at position (N-K)+1, we will get all windows.
                window_len = 1 + 2 * self.context_window
                shift_len = batch_size = window_len + 1
                # we need to make sure that we do not shift outside the boundaries of self.data too
                lastpos = data_len - 1  # index of the last word in data

                data_index = 0
                endpos = data_index + batch_size
                while True:
                    if endpos > lastpos:
                        break
                    currentTensor = data[data_index:endpos]
                    if len(currentTensor) < window_len:
                        break  # We are at the end
                    batch_x, batch_y = self.next_batch(currentTensor)
                    current_loss = self.run_optimization(batch_x, batch_y)
                    if step == 0 or step % 100 == 0:
                        logging.info("loss {}".format(current_loss))
                        loss_history.append(current_loss)
                    data_index += shift_len
                    endpos = data_index + batch_size
                    endpos = min(endpos,
                                 lastpos)  # takes care of last part of data. Maybe we should just ignore though
                    # Evaluation.
                    if do_display and (step % self.eval_step == 0 or step == 0):
                        self.display_words(x_test)
                    step += 1
                    self.run_optimization(batch_x, batch_y)
        return loss_history



class ContinuousBagOfWordsWord2Vec(Word2Vec):
    """Class to run word2vec using continuous bag of words (cbow).
    Attributes:
        word2id: A dictionary where the keys are nodes/words and values are integers that represent those nodes/words.
        id2word: A dictionary where the keys are integers and values are the nodes represented by the integers.
        data: A list or list of lists (if sentences or paths from node2vec).
        device_type: A string that indicates whether to run computations on (default=cpu).
        learning_rate: A float between 0 and 1 that controls how fast the model learns to solve the problem.
        batch_size: The size of each "batch" or slice of the data to sample when training the model.
        num_sepochs: The number of epochs to run when training the model (default=1).
        embedding_size: Dimension of embedded vectors.
        max_vocabulary_size: Maximum number of words (i.e. total number of different words in the vocabulary).
        min_occurrence: Minimum number of times a word needs to appear to be included (default=2).
        context_window: How many words to consider left and right.
        num_skips: How many times to reuse an input to generate a label.
        num_sampled: Number of negative examples to sample (default=7).
        display: An integer of the number of words to display.
        embedding: A 2D tensor with shape (samples, sequence_length), where each entry is a sequence of integers.
        batcher: A list of CBOW data for training; the first item is a batch and the second item is the batch labels.
        list_of_lists: A boolean which indicates whether or not the input data contains a list of lists.
        optimizer: The TensorFlow optimizer to use.
        data_index: An integer that stores the index of data for use when creating batches.
        current_sentence: An integer which is used to track the number of sentences or random walks.
        num_sentences: An integer that stores the total number of sentences.
        softmax_weights: A variable that stores the classifier weights.
        softmax_biases: A variable that stores classifier biases.
    Raises:
        TypeError: If the self.data does not contain a list of lists, where each list contains integers.
    """

    def __init__(self, data: List, worddictionary: Dict[str, int], reverse_worddictionary: Dict[int, str],
                 learning_rate: float = 0.1, batch_size: int = 128, num_epochs: int = 1, embedding_size: int = 200,
                 max_vocabulary_size: int = 50000, min_occurrence: int = 1, context_window: int = 3, num_skips: int = 2,
                 num_sampled: int = 7, display: Optional[int] = None, device_type: str = 'cpu') -> None:

        super().__init__(data=data, worddictionary=worddictionary, reverse_worddictionary=reverse_worddictionary,
                         learning_rate=learning_rate, batch_size=batch_size, num_epochs=num_epochs,
                         embedding_size=embedding_size, max_vocabulary_size=max_vocabulary_size,
                         min_occurrence=min_occurrence, context_window=context_window, num_skips=num_skips,
                         num_sampled=num_sampled, display=display, device_type=device_type)
        self.device_type = '/CPU:0' if 'cpu' in device_type.lower() else '/GPU:0'
        # set vocabulary size
        self.calculate_vocabulary_size()

        # with toy exs the # of nodes might be lower than the default value of num_sampled of 7. num_sampled needs to
        # be less than the # of exs (num_sampled is the # of negative samples that get evaluated per positive ex)
        if self.num_sampled > self.vocabulary_size:
            self.num_sampled = int(self.vocabulary_size / 2)

        self.optimizer: tf.keras.optimizers = tf.keras.optimizers.SGD(learning_rate)
        self.data_index: int = 0
        self.current_sentence: int = 0

        # do not display examples during training unless the user calls add_display_words (i.e. default is None)
        self.display = display
        self.n_epochs = num_epochs

        # ensure the following ops & var are assigned on CPU (some ops are not compatible on GPU)
        with tf.device(self.device_type):
            # create embedding (each row is a word embedding vector) with shape (#n_words, dims) and dim = vector size
            self.embedding: tf.Variable = tf.Variable(
                tf.random.uniform([self.vocabulary_size, embedding_size], -1.0, 1.0, dtype=tf.float32)
            )

            # should we initialize with uniform or normal?
            # # tf.Variable(tf.random.normal([self.vocabulary_size, embedding_size]))

            # construct the variables for the softmax loss
            tf_distribution = tf.random.truncated_normal([self.vocabulary_size, embedding_size],
                                                         stddev=0.5 / math.sqrt(embedding_size),
                                                         dtype=tf.float32)
            # get weights and biases
            self.softmax_weights: tf.Variable = tf.Variable(tf_distribution)
            self.softmax_biases = tf.Variable(tf.random.uniform([self.vocabulary_size], 0.0, 0.01))

    def cbow_embedding(self, x):
        """The function performs embedding lookups for each column in the input (except the middle one) and then
        averages the them to produce a word vector. The dimension of x is (batchsize, 2*context_window), e.g., (128,6).
        Note. x does not contain the middle word.
        Args:
            x: A batch-size long list of windows of words (sliding windows), for example:
                [[ 2619 15572 15573 15575 15576 15577], [15572 15573 15574 15576 15577 15578], ...]
        Returns:
            mean_embeddings: An averaged word embedding vector.

        Raises:
            ValueError: If the shape of stacked_embeddings is not equal to twice the context_window length.
        """

        stacked_embeddings = None
        # print('Defining %d embedding lookups representing each word in the context' % (2 * self.context_window))
        for i in range(2 * self.context_window):
            embedding_i = get_embedding(x[:, i], self.embedding)
            x_size, y_size = embedding_i.get_shape().as_list()  # added ',_' -- is this correct?

            if stacked_embeddings is None:
                stacked_embeddings = tf.reshape(embedding_i, [x_size, y_size, 1])
            else:
                stacked_embedding_value = [stacked_embeddings, tf.reshape(embedding_i, [x_size, y_size, 1])]
                stacked_embeddings = tf.concat(axis=2, values=stacked_embedding_value)

        assert stacked_embeddings.get_shape().as_list()[2] == 2 * self.context_window
        mean_embeddings = tf.reduce_mean(stacked_embeddings, 2, keepdims=False)

        return mean_embeddings

    def get_loss(self, mean_embeddings: tf.Tensor, y: np.ndarray) -> Union[float, int]:
        """Computes the softmax loss, using a sample of the negative labels each time. The inputs are embeddings of the
        train words with this loss we optimize weights, biases, embeddings.
        Args:
            mean_embeddings: A Tensor with shape [batch_size, dim].
            y: An array containing the target classes with shape [batch_size, num_true].
        Returns:
            loss: The softmax losses.
        """

        y = tf.cast(y, tf.int64)
        loss = tf.reduce_mean(tf.nn.sampled_softmax_loss(weights=self.softmax_weights,
                                                         biases=self.softmax_biases,
                                                         inputs=mean_embeddings,
                                                         labels=y,
                                                         num_sampled=self.num_sampled,
                                                         num_classes=self.vocabulary_size))

        return loss

    def nce_loss(self, x_embed: tf.Tensor, y: np.ndarray) -> Union[float, int]:
        """Calculates the noise-contrastive estimation (NCE) training loss estimation for each batch.
        Args:
            x_embed: A Tensor with shape [batch_size, dim].
            y: An array containing the target classes with shape [batch_size, num_true].
        Returns:
            loss: The NCE losses.
        """

        with tf.device(self.device_type):
            y = tf.cast(y, tf.int64)

            loss = tf.reduce_mean(tf.nn.nce_loss(weights=self.softmax_weights,
                                                 biases=self.softmax_biases,
                                                 labels=y,
                                                 inputs=x_embed,
                                                 num_sampled=self.num_sampled,
                                                 num_classes=self.vocabulary_size))

            return loss

    def evaluate(self, x_embed: tf.Tensor) -> tf.Tensor:
        """Computes the cosine similarity between a provided embedding and all other embedding vectors.
        Args:
            x_embed: A Tensor containing word embeddings.
        Returns:
            cosine_sim_op: A tensor of the cosine similarities between input data embedding and all other embeddings.
        """

        with tf.device(self.device_type):
            x_embed = tf.cast(x_embed, tf.float32)
            x_embed_norm = x_embed / tf.sqrt(tf.reduce_sum(tf.square(x_embed)))
            x_embed_sqrt = tf.sqrt(tf.reduce_sum(tf.square(self.embedding), 1, keepdims=True), tf.float32)
            embedding_norm = self.embedding / x_embed_sqrt

            # calculate cosine similarity
            cosine_sim_op = tf.matmul(x_embed_norm, embedding_norm, transpose_b=True)

            return cosine_sim_op

    def generate_batch_cbow(self, sentence: tf.Tensor) -> Tuple[np.ndarray, np.ndarray]:
        """Generates the next batch of data for CBOW.
        Args:
            sentence: A list of words to be used to create the batch
        Returns:
            A list whose first item is a batch and the second item is the batch's labels.
        """
        # span is the total size of the sliding window we look at [context_window central_word context_window]
        context_window = self.context_window
        span = 2 * context_window + 1
        # again, probably we can go: span = self.span
        sentencelen = len(sentence)
        batch_size = (sentencelen - (2 * context_window))
        # two numpy arrays to hold target (batch) and context words (labels). Batch has span-1=2*window_size columns
        batch = np.ndarray(shape=(batch_size, span - 1), dtype=np.int32)
        labels = np.ndarray(shape=(batch_size, 1), dtype=np.int64)
        sentence = sentence.numpy()
        # The buffer holds the data contained within the span and the deque essentially implements a sliding window
        buffer: collections.deque = collections.deque(maxlen=span)
        buffer.extend(sentence[0:span - 1])
        data_index = span - 1
        for i in range(batch_size):
            buffer.append(sentence[data_index])
            data_index += 1  # move sliding window 1 spot to the right
            context_words = [w for w in range(span) if w != context_window]
            for j, context_word in enumerate(context_words):
                batch[i, j] = buffer[context_word]
            labels[i, 0] = buffer[context_window]
        return batch, labels

    def run_optimization(self, x: np.array, y: np.array) -> None:
        """Runs optimization for each batch by retrieving an embedding and calculating loss. Once the loss has
        been calculated, the gradients are computed and the weights and biases are updated accordingly.
        Args:
            x: An array of integers to use as batch training data.
            y: An array of labels to use when evaluating loss for an epoch.
        Returns:
            None.
        """
        with tf.device(self.device_type):
            # wrap computation inside a GradientTape for automatic differentiation
            with tf.GradientTape() as g:
                emb = self.cbow_embedding(x)
                loss = self.nce_loss(emb, y)

            # compute gradients
            gradients = g.gradient(loss, [self.embedding, self.softmax_weights, self.softmax_biases])

            # Update W and b following gradients
            self.optimizer.apply_gradients(
                zip(gradients, [self.embedding, self.softmax_weights, self.softmax_biases]))
            return tf.reduce_sum(loss)

    def display_words(self) -> None:
        for w in self.display_examples:
            logging.info("{}: id={}".format(self.id2word[w], w))
        # x_test = np.array(self.display_examples)
        logging.info('Evaluation...\n')
        print("TODO -- fix me")
        if 2 + 2 != 5:
            return
        sim = calculate_cosine_similarity(self.cbow_embedding(x_test),  # type: ignore
                                          self.embedding,
                                          self.device_type).numpy()

        #print(sim[0])

        for i in range(len(self.display_examples)):
            top_k = 8  # number of nearest neighbors.
            nearest = (-sim[i, :]).argsort()[1:top_k + 1]
            disp_example = self.id2word[self.display_examples[i]]
            log_str = '{} nearest neighbors:'.format(disp_example)

            for k in range(top_k):
                log_str = '{} {},'.format(log_str, self.id2word[nearest[k]])

            #print(log_str)

        return None

    def train(self, display_step: int = 2000) -> List[None]:  # type: ignore
        """Trains a CBOW model.
        Args:
            display_step: An integer that is used to determine the number of steps to display when training the model.
        Returns:
            None.
        """
        # words for testing; display_step = 2000; eval_step = 2000
        if display_step is not None and 1 == 3:
            self.display_words()
        # data = self.data  #
        window_len = 2 * self.context_window + 1
        step = 0
        loss_history = []
        for _ in trange(1, self.n_epochs + 1):
            if self.list_of_lists or isinstance(self.data, tf.RaggedTensor):
                for sentence in self.data:
                    # Sentence is a Tensor
                    sentencelen = len(sentence)
                    if sentencelen < window_len:
                        continue
                    batch_x, batch_y = self.generate_batch_cbow(sentence)
                    current_loss = self.run_optimization(batch_x, batch_y)  # type: ignore
                    loss_history.append(current_loss)
                    if step % 100 == 0:
                        logging.info("loss {} ".format(current_loss))
                    step += 1
            else:
                data = self.data  #
                if not isinstance(data, tf.Tensor):
                    raise TypeError("We were expecting a Tensor object!")
                batch_size = self.batch_size
                data_len = len(data)
                # Note that we cannot fully digest all of the data in any one batch
                # if the window length is K and the natch_len is N, then the last
                # window that we get starts at position (N-K). Therefore, if we start
                # the next window at position (N-K)+1, we will get all windows.
                window_len = 1 + 2 * self.context_window
                shift_len = batch_size - window_len + 1
                # we need to make sure that we do not shift outside the boundaries of self.data too
                lastpos = data_len - 1  # index of the last word in data
                data_index = 0
                endpos = data_index + batch_size
                while endpos <= lastpos:
                    currentTensor = data[data_index:endpos]
                    if len(currentTensor) < window_len:
                        break  # We are at the end
                    batch_x, batch_y = self.next_batch(currentTensor)  # type: ignore
                    current_loss = self.run_optimization(batch_x, batch_y)  # type: ignore
                    if step == 0 or step % 100 == 0:
                        logging.info("loss {}".format(current_loss))
                        loss_history.append(current_loss)
                    data_index += shift_len
                    endpos = data_index + batch_size
                    endpos = min(endpos,
                                 lastpos)  # takes care of last part of data. Maybe we should just ignore though
                    # Evaluation.
            return loss_history



"""GraphTransformer class to convert graphs to edge embeddings."""
import numpy as np
from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module

from .edge_transformer import EdgeTransformer


class GraphTransformer:
    """GraphTransformer class to convert graphs to edge embeddings."""

    def __init__(self, method: str = "hadamard"):
        """Create new GraphTransformer object.

        Parameters
        ------------------------
        method: str = "hadamard",
            Method to use for the embedding.
            Can either be 'hadamard', 'average', 'weightedL1', 'weightedL2' or
            a custom lambda that receives two numpy arrays with the nodes
            embedding and returns the edge embedding.
        """
        self._transformer = EdgeTransformer(method=method)

    def fit(self, embedding: np.ndarray):
        """Fit the model.

        Parameters
        -------------------------
        embedding: np.ndarray,
            Embedding to use to fit the transformer.
        """
        self._transformer.fit(embedding)

    def transform(self, graph: EnsmallenGraph) -> np.ndarray:
        """Return edge embedding for given graph using provided method.

        Parameters
        --------------------------
        graph: EnsmallenGraph,
            The graph whose edges are to embed.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        return self._transformer.transform(graph.sources, graph.destinations)



"""NodeTransformer class to convert nodes to edge embeddings."""
import numpy as np


class NodeTransformer:
    """NodeTransformer class to convert nodes to edge embeddings."""

    def __init__(self):
        """Create new NodeTransformer object."""
        self._embedding = None

    def fit(self, embedding: np.ndarray):
        """Fit the model.

        Parameters
        -------------------------
        embedding: np.ndarray,
            Embedding to use to fit the transformer.
        """
        self._embedding = embedding

    def transform(self, nodes: np.ndarray) -> np.ndarray:
        """Return embeddings from given node.

        Parameters
        --------------------------
        nodes: np.ndarray,
            Vector of nodes whose embedding is to be returned.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        if self._embedding is None:
            raise ValueError(
                "Transformer was not fitted yet."
            )
        return self._embedding[nodes]



"""Module offers basic Corpus Transformer object, a simple class to tekenize textual corpuses."""
import string
from typing import Dict, List

import numpy as np
from nltk.corpus import stopwords
from nltk.corpus import wordnet as wn
from nltk.stem import PorterStemmer
from nltk.tokenize import word_tokenize
from tensorflow.keras.preprocessing.text import Tokenizer   # pylint: disable=import-error
from tqdm.auto import tqdm


class CorpusTransformer:
    """Simple class to tekenize textual corpuses."""

    def __init__(
        self,
        synonyms: Dict = None,
        language: str = "english",
        apply_stemming: bool = False,
        extend_synonyms: bool = True
    ):
        """Create new CorpusTransformer object.

        This is a GENERIC text tokenizer and is only useful for basic examples
        as in any advanced settings there will be need for a custom tokenizer.

        Parameters
        ----------------------------
        synonyms: Dict = None,
            The synonyms to use.
        language: str = "english",
            The language for the stopwords.
        apply_stemming: bool = False,
            Wethever to apply or not a stemming procedure, which
            by default is disabled.
            The algorithm used is a Porter Stemmer.
        extend_synonyms: bool = True,
            Wethever to automatically extend the synonyms using wordnet.
        """
        self._synonyms = {} if synonyms is None else synonyms
        self._stopwords = set(stopwords.words(
            language)) | set(string.punctuation)
        self._stemmer = PorterStemmer()
        self._apply_stemming = apply_stemming
        self._extend_synonyms = extend_synonyms
        self._tokenizer = None

    def get_synonym(self, word: str) -> str:
        """Return the synonym of the given word, if available.

        Parameters
        ----------------------------
        word: str,
            The word whose synonym is to be found.

        Returns
        ----------------------------
        The given word synonym.
        """
        if word not in self._synonyms:
            if not self._extend_synonyms:
                return word
            possible_synonyms = wn.synsets(word)
            if possible_synonyms:
                word_synonyms = [
                    w.lower()
                    for w in possible_synonyms[0].lemma_names()
                ]
                self._synonyms[word] = word_synonyms[0]
                self._synonyms[word_synonyms[0]] = word_synonyms[0]
            else:
                self._synonyms[word] = word

        return self._synonyms[word]

    def tokenize(self, texts: List[str], return_counts: bool = False, verbose: bool = True):
        """Fit model using stemming from given text.

        Parameters
        ----------------------------
        texts: List[str],
            The text to use to fit the transformer.
        return_counts: bool = False,
            Wethever to return the counts of the terms or not.
        verbose: bool = True,
            Wethever to show or not tokenization loading bar.

        Return
        -----------------------------
        Either the tokens or tuple containing the tokens and the counts.
        """
        all_tokens = []
        counter = {}
        for line in tqdm(texts, desc="Tokenizing", disable=not verbose):
            tokens = []
            for word in word_tokenize(line.lower()):
                if word not in self._stopwords:
                    synonym = self.get_synonym(word)
                    if self._apply_stemming:
                        synonym = self._stemmer.stem(synonym)
                    counter[synonym] = counter.setdefault(synonym, 0) + 1
                    tokens.append(synonym)
            all_tokens.append(tokens)

        if return_counts:
            return all_tokens, counter
        return all_tokens

    def fit(self, texts: List[str], min_count: int = 0, verbose: bool = True):
        """Fit the trasformer.

        Parameters
        ----------------------------
        texts: List[str],
            The texts to use for the fitting.
        min_count: int = 0,
            Minimum count to consider the word term.
        verbose: bool = True,
            Wethever to show or not the loading bars.
        """
        _, counts = self.tokenize(texts, True, verbose)

        self._stopwords |= {
            word
            for word, count in counts.items()
            if count <= min_count
        }

        self._tokenizer = Tokenizer()
        self._tokenizer.fit_on_texts((
            " ".join(tokens)
            for tokens in self.tokenize(texts, False, verbose)
        ))

    @property
    def vocabulary_size(self) -> int:
        """Return number of different terms."""
        return len(self._tokenizer.word_counts)

    def reverse_transform(self, sequences: List[List[int]]) -> List[str]:
        """Reverse the sequence to texts.

        Parameters
        ------------------------
        sequences: List[List[int]],
            The sequences to counter transform.

        Returns
        ------------------------
        The texts created from the given sequences.
        """
        return self._tokenizer.sequences_to_texts(sequences)

    def get_word_id(self, word: str) -> int:
        """Get the given words IDs.

        Parameters
        ------------------------
        word: int
            The word whose ID is to be retrieved.

        Returns
        ------------------------
        The word numeric ID.
        """
        return self._tokenizer.word_index[word]

    def transform(self, texts: List[str], min_length: int = 0, verbose: bool = True) -> np.ndarray:
        """Transform given text.

        Parameters
        --------------------------
        texts: List[str],
            The texts to encode as digits.
        min_length: int = 0,
            Minimum length of the single texts.
        verbose: bool = True,
            Wethever to show or not the loading bar.

        Returns
        --------------------------
        Numpy array with numpy arrays of tokens.
        """
        return np.array([
            np.array(tokens, dtype=np.int64) - 1
            for tokens in self._tokenizer.texts_to_sequences((
                " ".join(tokens)
                for tokens in self.tokenize(texts, verbose=verbose)
                if len(tokens) > min_length
            ))
        ])



"""GloVe model for graph and words embedding."""
from typing import Union

import tensorflow as tf
from tensorflow.keras import backend as K  # pylint: disable=import-error
from tensorflow.keras.layers import Add, Dot, Embedding, Flatten, Input  # pylint: disable=import-error
from tensorflow.keras.models import Model   # pylint: disable=import-error
from tensorflow.keras.optimizers import Optimizer   # pylint: disable=import-error

from .embedder import Embedder


class GloVe(Embedder):
    """GloVe model for graph and words embedding.

    The GloVe model for graoh embedding receives two words and is asked to
    predict its cooccurrence probability.
    """

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        optimizer: Union[str, Optimizer] = "nadam",
        alpha: float = 0.75
    ):
        """Create new GloVe-based Embedder object.

        Parameters
        ----------------------------
        vocabulary_size: int,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
        embedding_size: int,
            Dimension of the embedding.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
        alpha: float = 0.75,
            Alpha to use for the function
        """
        self._alpha = alpha
        super().__init__(
            vocabulary_size=vocabulary_size,
            embedding_size=embedding_size,
            optimizer=optimizer
        )

    def _glove_loss(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> float:
        """Compute the glove loss function.

        Parameters
        ---------------------------
        y_true: tf.Tensor,
            The true values Tensor for this batch.
        y_pred: tf.Tensor,
            The predicted values Tensor for this batch.

        Returns
        ---------------------------
        Loss function score related to this batch.
        """
        return K.sum(
            K.pow(K.clip(y_true, 0.0, 1.0), self._alpha) *
            K.square(y_pred - K.log(y_true)),
            axis=-1
        )

    def _build_model(self):
        """Create new Glove model."""
        # Creating the input layers
        input_layers = [
            Input((1,), name=Embedder.EMBEDDING_LAYER_NAME),
            Input((1,))
        ]

        # Creating the dot product of the embedding layers
        dot_product_layer = Dot(axes=2)([
            Embedding(
                self._vocabulary_size,
                self._embedding_size,
                input_length=1
            )(input_layer)
            for input_layer in input_layers
        ])

        # Creating the biases laye
        biases = [
            Embedding(
                self._vocabulary_size,
                1,
                input_length=1
            )(input_layer)
            for input_layer in input_layers
        ]

        # Concatenating with an add the three layers
        prediction = Flatten()(Add()([dot_product_layer, *biases]))

        # Creating the model
        glove = Model(
            inputs=input_layers,
            outputs=prediction,
            name="GloVe"
        )
        glove.compile(
            loss=self._glove_loss,
            optimizer=self._optimizer
        )

        return glove



"""Keras Sequence for running Neural Netwok on graph link prediction."""
from typing import Callable, Tuple, Union

import numpy as np
from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import Sequence

from ..transformers import EdgeTransformer


class LinkPredictionSequence(Sequence):
    """Keras Sequence for running Neural Netwok on graph link prediction."""

    def __init__(
        self,
        graph: EnsmallenGraph,
        embedding: np.ndarray,
        method: Union[str, Callable] = "hadamard",
        batch_size: int = 2**10,
        negative_samples: float = 1.0,
        graph_to_avoid: EnsmallenGraph = None,
        batches_per_epoch: bool = 2**8,
        avoid_self_loops: bool = False,
        elapsed_epochs: int = 0
    ):
        """Create new LinkPredictionSequence object.

        Parameters
        --------------------------------
        graph: EnsmallenGraph,
            The graph from which to sample the edges.
        embedding: np.ndarray,
            The embedding of the nodes.
        method: str = "hadamard",
            Method to use for the embedding.
            Can either be 'hadamard', 'average', 'weightedL1', 'weightedL2' or
            a custom lambda that receives two numpy arrays with the nodes
            embedding and returns the edge embedding.
        batch_size: int = 2**10,
            The batch size to use.
        negative_samples: float = 1.0,
            Factor of negatives to use in every batch.
            For example, with a batch size of 128 and negative_samples equal
            to 1.0, there will be 64 positives and 64 negatives.
        graph_to_avoid: EnsmallenGraph = None,
            Graph to avoid when generating the links.
            This can be the validation component of the graph, for example.
            More information to how to generate the holdouts is available
            in the EnsmallenGraph package.
        batches_per_epoch: bool = 2**8,
            Number of batches per epoch.
        avoid_self_loops: bool = False,
            If the self loops must be filtered away from the result.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        """
        self._graph = graph
        self._negative_samples = negative_samples
        self._graph_to_avoid = graph_to_avoid
        self._avoid_self_loops = avoid_self_loops
        self._transformer = EdgeTransformer(method)
        self._transformer.fit(embedding)
        super().__init__(
            samples_number=batches_per_epoch,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be rendered.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        edges, labels = self._graph.link_prediction(
            idx + self.elapsed_epochs,
            batch_size=self.batch_size,
            negative_samples=self._negative_samples,
            graph_to_avoid=self._graph_to_avoid,
            avoid_self_loops=self._avoid_self_loops
        )
        return (
            self._transformer.transform(
                edges[:, 0],
                edges[:, 1]
            ),
            labels
        )



"""Abstract Keras Sequence object for running models on graph walks."""
from typing import Dict

from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module
from .abstract_sequence import AbstractSequence


class AbstractNode2VecSequence(AbstractSequence):
    """Abstract Keras Sequence object for running models on graph walks."""

    def __init__(
        self,
        graph: EnsmallenGraph,
        walk_length: int,
        batch_size: int,
        iterations: int = 1,
        window_size: int = 4,
        shuffle: bool = True,
        min_length: int = 1,
        return_weight: float = 1.0,
        explore_weight: float = 1.0,
        change_node_type_weight: float = 1.0,
        change_edge_type_weight: float = 1.0,
        elapsed_epochs: int = 0,
        dense_nodes_mapping: Dict[int, int] = None
    ):
        """Create new Node2Vec Sequence object.

        Parameters
        -----------------------------
        graph: EnsmallenGraph,
            The graph from from where to extract the walks.
        walk_length: int,
            Maximal length of the walks.
            In directed graphs, when traps are present, walks may be shorter.
        batch_size: int,
            Number of nodes to include in a single batch.
        iterations: int = 1,
            Number of iterations of the single walks.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        shuffle: bool = True,
            Wthever to shuffle the vectors.
        min_length: int = 1,
            Minimum length of the walks.
            In directed graphs, when traps are present, walks shorter than
            this amount are removed. This should be two times the window_size.
        return_weight: float = 1.0,
            Weight on the probability of returning to node coming from
            Having this higher tends the walks to be
            more like a Breadth-First Search.
            Having this very high  (> 2) makes search very local.
            Equal to the inverse of p in the Node2Vec paper.
        explore_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node
            to the one we're coming from in the random walk
            Having this higher tends the walks to be
            more like a Depth-First Search.
            Having this very high makes search more outward.
            Having this very low makes search very local.
            Equal to the inverse of q in the Node2Vec paper.
        change_node_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node of a
            different type than the previous node. This only applies to
            colored graphs, otherwise it has no impact.
        change_edge_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor edge of a
            different type than the previous edge. This only applies to
            multigraphs, otherwise it has no impact.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        dense_nodes_mapping: Dict[int, int] = None,
            Mapping to use for converting sparse walk space into a dense space.
            This object can be created using the method available from graph
            called `get_dense_nodes_mapping` that returns a mapping from
            the non trap nodes (those from where a walk could start) and
            maps these nodes into a dense range of values.
        """
        self._graph = graph
        self._walk_length = walk_length
        self._iterations = iterations
        self._min_length = min_length
        self._return_weight = return_weight
        self._explore_weight = explore_weight
        self._change_node_type_weight = change_node_type_weight
        self._change_edge_type_weight = change_edge_type_weight
        self._dense_nodes_mapping = dense_nodes_mapping

        super().__init__(
            batch_size=batch_size,
            samples_number=self._graph.get_not_trap_nodes_number(),
            window_size=window_size,
            elapsed_epochs=elapsed_epochs
        )



"""Module with models for graph and text embedding and their Keras Sequences."""
from .embedders import CBOW, SkipGram, GloVe, BinarySkipGram
from .transformers import (
    NodeTransformer, EdgeTransformer, GraphTransformer, CorpusTransformer)
from .sequences import (NodeBinarySkipGramSequence, Node2VecSequence,
                        LinkPredictionSequence, WordBinarySkipGramSequence,
                        Word2VecSequence)
from .visualizations import GraphVisualizations

__all__ = [
    "CBOW", "SkipGram", "GloVe", "BinarySkipGram",
    "NodeBinarySkipGramSequence", "LinkPredictionSequence", "Node2VecSequence",
    "WordBinarySkipGramSequence", "Word2VecSequence",
    "NodeTransformer", "EdgeTransformer",
    "GraphTransformer", "CorpusTransformer",
    "GraphVisualizations"
]



"""Binary SkipGram model for graph and words embedding."""
from tensorflow.keras.layers import Dense, Dot, Embedding, Flatten, Input   # pylint: disable=import-error
from tensorflow.keras.metrics import AUC, Precision, Recall   # pylint: disable=import-error
from tensorflow.keras.models import Model  # pylint: disable=import-error

from .embedder import Embedder


class BinarySkipGram(Embedder):
    """Binary SkipGram model for graph and words embedding.

    The Binary SkipGram model receives a word and a candidate context
    and is asked to predict wethever if the context is actually or not
    in the context of the word. The prediction of this model is binary.
    """

    def _build_model(self):
        """Create new BinarySkipGram model."""

        # Creating the input layers
        input_layers = [
            Input(shape=(1,), name=Embedder.EMBEDDING_LAYER_NAME),
            Input(shape=(1,))
        ]

        # Creating dot product of embedding layers for each input layer
        dot_product_layer = Flatten()(Dot(axes=2)([
            Embedding(
                self._vocabulary_size,
                self._embedding_size,
                input_length=1
            )(input_layer)
            for input_layer in input_layers
        ]))

        # Output layer from previous dot product
        output = Dense(1, activation="sigmoid")(dot_product_layer)

        # Creating the actual model
        skipgram = Model(
            inputs=input_layers,
            outputs=output,
            name="BinarySkipGram"
        )

        # Compiling the model
        skipgram.compile(
            loss='binary_crossentropy',
            optimizer=self._optimizer,
            metrics=[
                AUC(curve="ROC", name="auroc"),
                AUC(curve="PR", name="auprc"),
                Precision(),
                Recall()
            ]
        )

        return skipgram



"""Module with graph embedding models."""
from .glove import GloVe
from .skipgram import SkipGram
from .binary_skipgram import BinarySkipGram
from .cbow import CBOW

__all__ = [
    "GloVe", "SkipGram", "CBOW", "BinarySkipGram"
]



"""Abstract Keras Sequence object for running models on graph walks."""
from typing import List

import numpy as np  # type: ignore
from keras_mixed_sequence import VectorSequence

from .abstract_sequence import AbstractSequence


class AbstractWord2VecSequence(AbstractSequence):
    """Abstract Keras Sequence object for running models on texts."""

    def __init__(
        self,
        sequences: List[np.ndarray],
        batch_size: int,
        window_size: int = 4,
        shuffle: bool = True,
        seed: int = 42,
        elapsed_epochs: int = 0,
    ):
        """Create new Node2Vec Sequence object.

        Parameters
        -----------------------------
        sequences: List[np.ndarray],
            List of sequences of integers.
        batch_size: int,
            Number of nodes to include in a single batch.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        shuffle: bool = True,
            Wthever to shuffle the vectors.
        seed: int = 42,
            The seed to use to make extraction reproducible.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        """

        self._sequences = VectorSequence(
            sequences,
            batch_size,
            seed=seed,
            elapsed_epochs=elapsed_epochs
        )
        super().__init__(
            window_size=window_size,
            shuffle=shuffle,
            samples_number=self._sequences.samples_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )

    def on_epoch_end(self):
        """Shuffles given sequences object."""
        super().on_epoch_end()
        self._sequences.on_epoch_end()



"""Module with Keras Sequences."""
from .node2vec_sequence import Node2VecSequence
from .node_binary_skipgram_sequence import NodeBinarySkipGramSequence
from .link_prediction_sequence import LinkPredictionSequence
from .word2vec import Word2VecSequence
from .word_binary_skipgram_sequence import WordBinarySkipGramSequence

__all__ = [
    "Node2VecSequence", "NodeBinarySkipGramSequence",
    "LinkPredictionSequence", "Word2VecSequence",
    "WordBinarySkipGramSequence"
]
